/*
 * Copyright (C) 2013 by the University of Southern California
 * $Id: server.cc 20047 2014-02-27 12:39:47Z liangzhu $
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License,
 * version 2, as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
 *
 */

#include "server.hh"
#include <linux/tcp.h>

//------------------------------------------
//global var
//------------------------------------------
vector<int> sockets;
int dns_socket_tcp = -1;
int dns_socket_udp = -1;

//tls
bool done_tls_init = false;
pthread_mutex_t check_tls_mutex;	//lock

#ifdef USE_GNUTLS
gnutls_session_t tls_session = NULL;
#endif

#ifdef USE_OPENSSL
SSL_CTX* ctx = NULL;
SSL* ssl = NULL;
#endif

#define KEYFILE "key.pem"
#define CERTFILE "cert.pem"
#define CAFILE "ca-bundle.trust.crt"
#define CRLFILE "crl.pem"

//pthread_cond_t tls_session_cv;

//query table and related lock and condition var
//int query_in_process = 0;		//# of queries sent but not responsed
pthread_cond_t query_table_cv;	//CV	// = PTHREAD_COND_INITIALIZER;
map<int, string> query_table;		//query table
pthread_mutex_t query_table_mutex;	//lock	// = PTHREAD_MUTEX_INITIALIZER;
pthread_t *recv_resp_thread = NULL;	//response thread
pthread_t *send_query_thread = NULL;		//send query thread
map<pthread_t, string> thread_map;	//thread_map used to find thread name

//received queries from client, pending to send to server
list<net_msg_t*> pending_query;
pthread_mutex_t pending_query_mutex;
pthread_cond_t pending_query_cv;

map<int, dns_ident_t> udp_clnt_map;
pthread_mutex_t udp_clnt_map_mutex;
list<u_int16_t> proxy_id_list;

//-------------------------------------------
//end of global var
//-------------------------------------------

/*
	argument passed to thread
*/
struct thread_argv_t {
	int s;	//socket
	//gnutls_session_t session;	//tls session
	int t;	//type
	string name;//thread name
	string pname;//process name
	void *value;
	thread_argv_t() {
		s = -1;
	//	session = NULL;
		t = CONN_UNKNOWN;
		name = "UNKNOWN";
		value = NULL;
	}
};

/*
	clean up all the sockets
*/
void clean_all_skt()
{
	cerr << "[info] clean up all the sockets."<<endl;
        for (int i=0; i<sockets.size(); i++) {
        	shutdown (sockets[i], SHUT_RDWR);     /* no more receptions */
        	close(sockets[i]);
        }	
}

/*
	handler for thread
*/
void sigthread_handler (int s) 
{
	cerr << "[info] sig thread handler: SIGALRM, exit current thread..." << endl;
	pthread_t self_id = pthread_self();
	fprintf(stderr, "[info] thread: %s (%u) received SIGALRM, ending\n", thread_map[self_id].c_str(), self_id);	
	pthread_exit(NULL);
}


/*
	handler for the main process
*/
void sig_handler (int s)
{
	if (s == SIGINT) {
		cerr << "[info] user kills the program. (ctrl-c)" << endl;
	} else if (s == SIGQUIT) {
		cerr << "[info] Terminate process and generate core dump." << endl;
	}
	clean_all_skt();
        exit(0);
}

/*
	clean up child process: fork
*/
void sigchld_handler(int s)
{
	while(waitpid(-1, NULL, WNOHANG) > 0);
}

#ifdef USE_GNUTLS
/*
this function is copied from 
http://gnutls.org/manual/html_node/Echo-server-with-X_002e509-authentication.html#Echo-server-with-X_002e509-authentication
*/
static gnutls_dh_params_t dh_params;

static int generate_dh_params(void)
{
        unsigned int bits = gnutls_sec_param_to_pk_bits(GNUTLS_PK_DH,
                                                        GNUTLS_SEC_PARAM_LEGACY);

        /* Generate Diffie-Hellman parameters - for use with DHE
         * kx algorithms. When short bit length is used, it might
         * be wise to regenerate parameters often.
         */
        gnutls_dh_params_init(&dh_params);
        gnutls_dh_params_generate2(dh_params, bits);

        return 0;
}
#endif //#ifdef USE_GNUTLS

void handle_tcp_client(int s)
{
#ifdef USE_GNUTLS
	gnutls_session_t session;
	gnutls_anon_server_credentials_t anoncred;
#endif //#ifdef USE_GNUTLS
	int ret = -1;
	int send_len = -1;
	int recv_len = -1;
	int query_size = -1;
	int ans_size = -1;
	unsigned char* query_pkt = NULL;
	unsigned char* ans_pkt = NULL;	
	bool up_tls = false;
	bool first_packet = true;
	
	struct sockaddr client_addr;
	socklen_t client_addr_len = sizeof(client_addr);
	struct sockaddr serv_addr;
	socklen_t serv_addr_len = sizeof(serv_addr);
	
	unsigned char buf[MIN_BUF_SIZE];
	memset(buf, 0, MIN_BUF_SIZE);
	string msg_tls_fine = MSG_TLS_OK;
	memcpy ( buf, msg_tls_fine.c_str(), msg_tls_fine.length() + 1);
	
	do {
		if (!up_tls) 
			query_pkt = tcp_pull_dns_pkt(s, query_size);
		else {
#ifdef USE_GNUTLS
			query_pkt = tls_pull_dns_pkt(session, query_size);
#endif //#ifdef USE_GNUTLS
		}
		print_dns_msg(query_pkt + 2, query_size-2);
		LOG( LOG_DEBUG, "[DEBUG] read a DNS packet from the socket:%d\n",s);
		LOG( LOG_DEBUG, "[DEBUG] this is %s first packet from socket:%d\n",(first_packet ? "" : "not"), s);

		if (first_packet) {
			first_packet = false;
			if (is_tls_set (query_pkt + 2, query_size-2)) {//tls initialization
				up_tls = true;				
				LOG( LOG_DEBUG, "tls bit is set with fake query. going to tls init\n");
				if (!done_tls_init) {
					done_tls_init = true;
#ifdef USE_GNUTLS
					//gnutls_global_init ();
					gnutls_anon_allocate_server_credentials (&anoncred);
					//generate_dh_params ();
					gnutls_anon_set_server_dh_params (anoncred, dh_params);
					gnutls_init (&session, GNUTLS_SERVER);
				//gnutls_priority_set_direct (session, "NORMAL:+ANON-ECDH:+ANON-DH", NULL);
					gnutls_priority_set_direct(session, "PERFORMANCE:+ANON-DH:!ARCFOUR-128", NULL);
					gnutls_credentials_set (session, GNUTLS_CRD_ANON, anoncred);
					gnutls_transport_set_int (session, s);
					
					//send tls ok back to client
					send_len = sendpkt(s, buf, msg_tls_fine.length() + 1);					
					
					do {
						ret = gnutls_handshake (session);
					} while (ret < 0 && gnutls_error_is_fatal (ret) == 0);
					if (ret < 0){
						shutdown (s, SHUT_RDWR); 
						close (s);
						gnutls_deinit (session);
						fprintf (stderr, "*** Handshake has failed (%s)\n\n",gnutls_strerror (ret));
			  			return;
			  		}
#endif //#ifdef USE_GNUTLS
			  		printf ("- Handshake was completed\n");
				}				
				continue;
			}
		}

		if (dns_socket_tcp == -1) {
			dns_socket_tcp = tcp_connect (DNS_SERV_PORT, DNS_SERV_IP);//this is the real server
			LOG( LOG_DEBUG, "[DEBUG] tcp connect to DNS server... dns_socket:%d\n",dns_socket_tcp);		
		}
		send_len = sendpkt(dns_socket_tcp, query_pkt, query_size);					
			//cout<<"send_len="<<send_len<<endl;
		while (send_len == -1) {//this has to be successful
			dns_socket_tcp = tcp_connect (DNS_SERV_PORT, DNS_SERV_IP);//this is the real server	
			send_len = sendpkt(dns_socket_tcp, query_pkt, query_size);
			LOG( LOG_DEBUG, "[DEBUG] dns_socket_tcp:%d is dead, connect to DNS server again\n",dns_socket_tcp);
		}
		ans_pkt = tcp_pull_dns_pkt(dns_socket_tcp, ans_size);
		LOG( LOG_DEBUG, "[DEBUG] receive the ans from dns_socket_tcp:%d\n",dns_socket_tcp);	
		print_dns_msg(ans_pkt + 2, ans_size-2);
			
		//send back to client
		if (!up_tls) 
			send_len = sendpkt(s, ans_pkt, ans_size);
		else {
#ifdef USE_GNUTLS
			send_len = gnutls_record_send (session, ans_pkt, ans_size);
#endif //#ifdef USE_GNUTLS
		}
		if (send_len == -1) die ("send returns -1");
		if (query_pkt != NULL) 
			delete[] query_pkt;
		if (ans_pkt != NULL) 
			delete[] ans_pkt;
	
	} while (query_pkt != NULL);
	
	if (up_tls) {
#ifdef USE_GNUTLS
		gnutls_bye (session, GNUTLS_SHUT_WR);
		gnutls_deinit (session);
		gnutls_anon_free_server_credentials (anoncred);
		gnutls_global_deinit ();
#endif //#ifdef USE_GNUTLS
	}
	shutdown (s, SHUT_RDWR); 
	close(s); 
}

/*
	clean up packet, delect the memory and reset the pointer
*/
void cleanup_pkt (unsigned char* pkt)
{
	if (pkt != NULL) {
		delete[] pkt;
		pkt = NULL;
	}		
}

/*
	resv queries from client and store them in query_table
*/
void *recv_query_from_client(void *func_argv)
{
	thread_argv_t *my_arg = (thread_argv_t *)func_argv;
	string thread_name = my_arg->name;
	string pname = my_arg->pname;
	const char *myname = pname.c_str();
	const char *tname = thread_name.c_str();
	LOG( LOG_DEBUG, "[DEBUG] %s thread[%s] starts\n",myname, tname);	
	int s = my_arg->s;

#ifdef USE_GNUTLS
	gnutls_certificate_credentials_t x509_cred;
	gnutls_priority_t priority_cache;
#endif //USE_GNUTLS

	int ret = -1;
	int recv_len = -1;
	int send_len = -1;
	int query_size = -1;
	unsigned char* query_pkt = NULL;	
	bool first_packet = true;
	//bool up_tls = false;
	
	unsigned char buf[MIN_BUF_SIZE];
	memset(buf, 0, MIN_BUF_SIZE);
	string msg_tls_fine = MSG_TLS_OK;
	memcpy ( buf, msg_tls_fine.c_str(), msg_tls_fine.length() + 1);
	net_msg_t *dns_msg = NULL;
	
	struct sockaddr clnt_addr;
	socklen_t clnt_addr_len = sizeof(clnt_addr);

	unsigned char ssl_buf[MAX_BUF_SIZE];
	unsigned char recv_buf[MAX_BUF_SIZE];
	uint16_t len;
	int ssl_read_len;
	while (1) {
		LOG( LOG_DEBUG, "[DEBUG] %s thread[%s] %s waiting for a dns packet...\n", myname, tname,
		(my_arg->t == CONN_TLS ? "TLS":(my_arg->t == CONN_TCP? "TCP":"UDP"))
		);
		if (my_arg->t == CONN_TCP) {
			len = 0;
			memset(recv_buf, 0, MAX_BUF_SIZE);
			recv_len = recv(s, (void*)&len, sizeof(len), 0);
			if (recv_len == 0) { //close tls
				query_pkt = NULL;
			} else if (recv_len < (ssize_t)sizeof(len)) {
				//cout << "recv_len =" << recv_len << endl;
				perror("recv"); 
				die("tcp recv< sizeof(len)");
			} else {
				len = ntohs(len);
				if (recv(s, recv_buf, len, 0) < (ssize_t)len) {
					die("tcp recv<0");
				}
				query_size = len;
				query_pkt = recv_buf;			
			}
			len = ntohs(len);
			//query_pkt = tcp_pull_dns_pkt(s, query_size);
		} else if (my_arg->t == CONN_TLS) {//this set by this thread, so tls_session != NULL at this point
			//pthread_mutex_lock(&check_tls_mutex);
			//if (tls_session == NULL);
			//pthread_mutex_unlock(&check_tls_mutex);
			len = 0;
			memset(ssl_buf, 0, MAX_BUF_SIZE);
#ifdef USE_GNUTLS	 
			//query_pkt = tls_pull_dns_pkt(tls_session, query_size);
			ssl_read_len = gnutls_record_recv(tls_session, (void*)&len, (int)sizeof(len));
			if (ssl_read_len == 0) {//close tls
				query_pkt = NULL;
			} else if (ssl_read_len < 0) {
				die("gnutls_record_recv<0");
			} else {
				len = ntohs(len);
				if(gnutls_record_recv(tls_session, ssl_buf, len) < 0) {
					die("gnutls_record_recv<0");
				}
				query_size = len;
				query_pkt = ssl_buf;
			}
#endif// USE_GNUTLS
#ifdef USE_OPENSSL
			ssl_read_len = SSL_read(ssl, (void*)&len, (int)sizeof(len));
			if (ssl_read_len == 0) {//close tls
				query_pkt = NULL;
			} else if (ssl_read_len < 0) {
				log_crypto_err("could not SSL_read");
				exit(-1);
			} else {
				len = ntohs(len);
				if(SSL_read(ssl, ssl_buf, len) < 0) {
					log_crypto_err("could not SSL_read");
					exit(-1);
				}
				query_size = len;
				query_pkt = ssl_buf;
			}
#endif
		} else if (my_arg->t == CONN_UDP) {
			query_pkt = udp_pull_dns_pkt(s, query_size, &clnt_addr, &clnt_addr_len);
		} else {
			die ("thread[send_query] client connection type is unkonw");
		}
		if (query_pkt == NULL) {
			if (my_arg->t == CONN_TLS) {
				LOG( LOG_DEBUG, "[DEBUG] %s thread[%s] TLS reading pkt==NULL, client closes TLS, clean up TLS, break, will clean up tcp...\n", myname, tname);
#ifdef USE_GNUTLS
				gnutls_bye (tls_session, GNUTLS_SHUT_WR);
				gnutls_deinit (tls_session);
				gnutls_certificate_free_credentials(x509_cred);
				gnutls_priority_deinit(priority_cache);
#endif //#ifdef USE_GNUTLS
#ifdef USE_OPENSSL
				SSL_shutdown(ssl);
				SSL_free(ssl);
				SSL_CTX_free(ctx);
#endif
				LOG( LOG_DEBUG, "[DEBUG] %s thread[%s] done TLS cleanup\n",myname, tname);
				my_arg->t = CONN_TCP;
				cleanup_pkt(query_pkt);						
				break;			
				//continue;
			} else if (my_arg->t == CONN_TCP) {
				LOG( LOG_DEBUG, "[DEBUG] %s thread[%s] TCP reading pkt==NULL, client closes TCP, break, will clean up\n",myname, tname);
				break;	
			} else if (my_arg->t == CONN_UDP) {
				LOG( LOG_DEBUG, "[DEBUG] %s thread[%s] UDP reading pkt==NULL, client closes UDP, break, will clean up\n",myname, tname);
				break;
			}
		}
		//u_int16_t dns_id = 0;	
		//make clean packet information for UDP, since we only send UDP to server
		//if (my_arg->t == CONN_TCP) {
		//	query_pkt += 2;
		//	query_size -= 2;
			//dns_id = get_dns_id(ans_pkt + 2, ans_size - 2);
		//} //else {
			
		//	dns_id = get_dns_id(ans_pkt, ans_size);
		//}
		LOG( LOG_DEBUG, "[DEBUG] %s thread[%s] read a DNS packet from the socket:%d\n",myname, tname,s);
		print_dns_msg(query_pkt, query_size);
		LOG( LOG_DEBUG, "[DEBUG] %s thread[%s] this is %s first packet from socket:%d\n",myname, tname,(first_packet ? "" : "not "), s);

		if (first_packet) {
			first_packet = false;
			/*if (is_tls_set (
					( my_arg->t == CONN_UDP ? query_pkt : (query_pkt + 2) ), 
					( my_arg->t == CONN_UDP ? query_size : (query_size - 2))
					)
			)*/
			if (is_tls_set(query_pkt, query_size)) {//tls initialization
				//up_tls = true;
				my_arg->t = CONN_TLS;
				LOG( LOG_DEBUG, "[DEBUG] %s thread[%s] tls bit is set with fake query. going to tls init\n",myname, tname);
				pthread_mutex_lock(&check_tls_mutex);	
				if (!done_tls_init) {
					done_tls_init = true;

#ifdef USE_GNUTLS
					gnutls_certificate_allocate_credentials(&x509_cred);
					gnutls_certificate_set_x509_trust_file(x509_cred, CAFILE, GNUTLS_X509_FMT_PEM);
					gnutls_certificate_set_x509_crl_file(x509_cred, CRLFILE, GNUTLS_X509_FMT_PEM);
					ret = gnutls_certificate_set_x509_key_file(x509_cred, CERTFILE, KEYFILE, GNUTLS_X509_FMT_PEM);
					if (ret < 0) {
						die("No certificate or key were found");
					}
					LOG( LOG_DEBUG, "[DEBUG] %s thread[%s] loading cert done\n",myname, tname);
					gnutls_certificate_set_dh_params(x509_cred, dh_params);
					gnutls_priority_init(&priority_cache,"PERFORMANCE:%SERVER_PRECEDENCE", NULL);
					gnutls_init (&tls_session, GNUTLS_SERVER);
					gnutls_priority_set(tls_session, priority_cache);
					gnutls_credentials_set(tls_session, GNUTLS_CRD_CERTIFICATE, x509_cred);
					gnutls_certificate_server_set_request(tls_session, GNUTLS_CERT_IGNORE);
					gnutls_transport_set_int (tls_session, s);
#endif //#ifdef USE_GNUTLS

#ifdef USE_OPENSSL
					LOG( LOG_DEBUG, "[DEBUG] %s thread[%s] create ssl structure\n",myname, tname);
					ctx = SSL_CTX_new (SSLv23_server_method ());
					if (!ctx) ERR_print_errors_fp (stderr);
					LOG( LOG_DEBUG, "[DEBUG] %s thread[%s] loading certificate\n",myname, tname);
					if ( SSL_CTX_use_certificate_file(ctx, CERTFILE, SSL_FILETYPE_PEM) <= 0 ) {
						ERR_print_errors_fp(stderr);
						die("cannot load cerificate file");
					}
					if ( SSL_CTX_use_PrivateKey_file(ctx, KEYFILE, SSL_FILETYPE_PEM) <= 0 ) {
						ERR_print_errors_fp(stderr);
						die("cannot load key file");
					}
					if ( !SSL_CTX_check_private_key(ctx) ) {
						//fprintf(stderr, "Private key does not match the public certificate\n");
						die("Private key does not match the public certificate");
					}
					LOG( LOG_DEBUG, "[DEBUG] %s thread[%s] set ssl socket\n",myname, tname);
					ssl = (SSL*)incoming_ssl_fd(ctx, s);
					if(!ssl) {
						ERR_print_errors_fp (stderr);
						die("cannot create ssl");
					}
#endif //USE_OPENSSL

					LOG( LOG_DEBUG, "[DEBUG] %s thread[%s] sending tls ok back to client\n",myname, tname);		
					//need to send back sth
					send_len = sendpkt(s, query_pkt - 2, query_size + 2);					

#ifdef USE_OPENSSL
					LOG( LOG_DEBUG, "[DEBUG] %s thread[%s] ssl accept\n",myname, tname);
					if ( SSL_accept(ssl) == -1 ) {
        					ERR_print_errors_fp(stderr);
						die("ssl accept fails");
					}
					LOG( LOG_DEBUG, "[DEBUG] %s thread[%s] done ssl accept\n",myname, tname);
#endif //USE_OPENSSL

#ifdef USE_GNUTLS					
					do {
						ret = gnutls_handshake (tls_session);
					} while (ret < 0 && gnutls_error_is_fatal (ret) == 0);
					if (ret < 0){
						shutdown (s, SHUT_RDWR); 
						close (s);
						gnutls_deinit (tls_session);
						fprintf (stderr, "*** thread[send_query] Handshake has failed (%s)\n\n",gnutls_strerror (ret));
			  			return NULL;
			  		}
#endif //#ifdef USE_GNUTLS
			  		printf ("- thread[send_query] Handshake was completed\n");
				}
				pthread_mutex_unlock(&check_tls_mutex);					
				continue;
			}
		}
		
		dns_msg = new net_msg_t();
		dns_msg->copy(query_pkt, query_size);
		if (my_arg->t == CONN_UDP) {
			memcpy(&(dns_msg->clnt_addr), &clnt_addr, clnt_addr_len);
			dns_msg->clnt_addr_len = clnt_addr_len;		
		}
		//net_msg_t *dns_msg = new net_msg_t(query_pkt, query_size);
		
		//add the query to pending_query
		pthread_mutex_lock(&pending_query_mutex);		
		pending_query.push_back(dns_msg);
		if (my_arg->t == CONN_TCP) {//make it original for delelte[], otehrwise mem leak
			query_pkt -= 2;
		}
		//cout << "before"<<endl;
		if (my_arg->t == CONN_UDP) {//tls/tls use fix buff
			cleanup_pkt(query_pkt);
		}
		//cout << "after" <<endl;
		LOG( LOG_DEBUG, "[DEBUG] %s thread[%s] put 1 query into pending query (size:%d) signal: pending_query_cv\n",myname, tname, pending_query.size());
		pthread_cond_signal(&pending_query_cv);
		pthread_mutex_unlock(&pending_query_mutex);
	}
	LOG( LOG_DEBUG, "[DEBUG] %s thread[%s] ending...\n",myname, tname);
	LOG( LOG_DEBUG, "[DEBUG] %s thread[%s] sends alarm to recv_resp_thread and send_query_thread\n",myname, tname);
	int rc;
	rc = pthread_kill(*recv_resp_thread, SIGALRM); //tell the recv thread to exit
	rc = pthread_kill(*send_query_thread, SIGALRM); //tell the recv thread to exit
	
	/*if (my_arg->t == CONN_TLS) {
		//pthread_mutex_lock(&check_tls_mutex);//lock
		gnutls_bye (tls_session, GNUTLS_SHUT_WR);
		gnutls_deinit (tls_session);
		gnutls_anon_free_server_credentials (anoncred);
		gnutls_global_deinit ();
		//pthread_mutex_unlock(&check_tls_mutex);//unlock
	}*/
	close(dns_socket_udp);
	pthread_exit(NULL);	
}


void *send_query_to_server(void *func_argv)
{
	thread_argv_t *my_arg = (thread_argv_t *)func_argv;
	string thread_name = my_arg->name;
	int send_len = -1;
	u_int16_t dns_id = 0;
	u_int16_t dns_id_tmp = 0;
	string pname = my_arg->pname;
	const char *myname = pname.c_str();
	const char *tname = thread_name.c_str();
	LOG( LOG_DEBUG, "[DEBUG] %s thread[%s] starts\n",myname, tname);
	if (dns_socket_udp == -1) {
		dns_socket_udp = udp_connect (DNS_SERV_PORT, DNS_SERV_IP);//this is the real server
		LOG( LOG_DEBUG, "[DEBUG] %s thread[%s] udp connect to DNS server... dns_socket_udp:%d\n",myname, tname, dns_socket_udp);		
	}
	net_msg_t *dns_msg = NULL;
	while (1) {
		pthread_mutex_lock(&pending_query_mutex);	//pending query lock
		while (pending_query.size() == 0) {		//no pending query
			LOG( LOG_DEBUG, "[DEBUG] %s thread[%s] no pending queries: wait pending_query_cv\n",myname, tname);
			pthread_cond_wait(&pending_query_cv, &pending_query_mutex);	//wait for pending query
			LOG( LOG_DEBUG, "[DEBUG] %s thread[%s] no pending queries: wake up from pending_query_cv\n",myname, tname);
		}
		//wake up, there must be at least pending query
		pthread_mutex_lock(&query_table_mutex);		//query table lock
		while (query_table.size() >= MAX_NUM_QUERY) {	//too much queries, cannot sent now
			LOG( LOG_DEBUG, "[DEBUG] %s thread[%s]  too much queries:%d wait query_table_cv\n",myname, tname, query_table.size());
			pthread_mutex_unlock(&pending_query_mutex);//release lock, pending query only be more and more
			pthread_cond_wait(&query_table_cv, &query_table_mutex);
			//wake up, now can sent query, there is no need to check pending query because there must be at least one at this point
			LOG( LOG_DEBUG, "[DEBUG] %s thread[%s]  wake up from query_table_cv\n",myname, tname);
			pthread_mutex_lock(&pending_query_mutex);	//pending query lock
		}
		LOG( LOG_DEBUG, "[DEBUG] %s thread[%s] # of pending queries:%d\n",myname, tname,pending_query.size());
		LOG( LOG_DEBUG, "[DEBUG] %s thread[%s] pop 1 query from pending queries and sending to server\n",myname, tname);
		dns_msg = pending_query.front();
		//printf("0x%x\n", dns_msg->msg);
		pending_query.pop_front();	//THIS WILL NOT destroy the poped object, because it's pointer
		//printf("0x%x\n", dns_msg->msg);
		LOG( LOG_DEBUG, "[DEBUG] %s thread[%s] sending a pkt to server, msg_len=%d\n",myname, tname, dns_msg->len);

		dns_id = get_dns_id(dns_msg->msg, dns_msg->len);
		//to save one lock, proxy_id_list is locked by query_table_mutex which is a little strange
		if (my_arg->t == CONN_UDP) {
			//pthread_mutex_lock(&udp_clnt_map_mutex);
			if (proxy_id_list.size() == 0) {
				die("proxy runs out of id... exit!");
			}
			dns_id_tmp = proxy_id_list.front();
			proxy_id_list.pop_front();		
			
			dns_ident_t dns_ident;
			memcpy(&(dns_ident.clnt_addr), &(dns_msg->clnt_addr), dns_msg->clnt_addr_len);//copy client address
			dns_ident.clnt_addr_len = dns_msg->clnt_addr_len;
			dns_ident.dns_id = dns_id;//copy orignal dns id
			
			put_dns_id(dns_msg->msg, dns_id_tmp);//change dns id in dns message
			udp_clnt_map[dns_id_tmp] = dns_ident;//log it in record
			LOG( LOG_DEBUG, "[DEBUG] %s thread[%s] done changing dns_id, id_before:%d, id_after:%d\n",myname, tname, dns_id, dns_id_tmp);	
			dns_id = dns_id_tmp;
			//pthread_mutex_unlock(&udp_clnt_map_mutex);
			
			if (get_dns_id(dns_msg->msg, dns_msg->len) != dns_id_tmp) {
				cerr << "[ERROR] put dns id != get dns id" << endl;
				exit(-1);
			}
		}
		send_len = sendpkt(dns_socket_udp, dns_msg->msg, dns_msg->len);
		if (send_len == -1) die ("thread[send_query_to_server] send returns -1");
		//dns_id = get_dns_id(dns_msg->msg, dns_msg->len);
		query_table[dns_id] = "sent";
		
		//NOTE that actually we don't need to call following two statement because pop_front has already done this
		//printf("0x%x\n", dns_msg->msg);
		//delete[] dns_msg->msg;
		//cleanup_pkt(dns_msg->msg);
		if (dns_msg != NULL) {
			delete dns_msg;
			dns_msg = NULL;
		}
		pthread_mutex_unlock(&query_table_mutex);
		pthread_mutex_unlock(&pending_query_mutex);
	}
}

/*
	recv response from dns server and send back to client
*/
void *recv_resp_from_server(void *func_argv)
{
	thread_argv_t *my_arg = (thread_argv_t *)func_argv;
	string thread_name = my_arg->name;
	string pname = my_arg->pname;
	const char *myname = pname.c_str();
	const char *tname = thread_name.c_str();
	LOG( LOG_DEBUG, "[DEBUG] %s thread[%s] starts\n", myname, tname);

	int send_len = -1;
	int recv_len = -1;
	int ans_size = -1;
	unsigned char* ans_pkt = NULL;
	unsigned char *tcp_pkt = NULL;
	
	bool need_check_tls = true;
	
	//struct sockaddr client_addr;
	//socklen_t client_addr_len = sizeof(client_addr);
	struct sockaddr serv_addr;
	socklen_t serv_addr_len = sizeof(serv_addr);
	u_int16_t dns_id = 0;
	dns_ident_t *dns_ident = NULL;
	
	while ((ans_pkt = udp_pull_dns_pkt(dns_socket_udp, ans_size, &serv_addr, &serv_addr_len)) != NULL) {
		LOG( LOG_DEBUG, "[DEBUG] %s thread[%s] receive a ans from dns_socket_udp:%d\n", myname, tname,dns_socket_udp);
		print_dns_msg(ans_pkt, ans_size);
		dns_id = get_dns_id(ans_pkt, ans_size);
		
		pthread_mutex_lock(&check_tls_mutex);//lock
		if (done_tls_init)
			my_arg->t = CONN_TLS;
		pthread_mutex_unlock(&check_tls_mutex);//unlock
		
		//send ans back to client
		if (my_arg->t == CONN_TLS) {//tls
			tcp_pkt = make_tcp_pkt(ans_pkt, ans_size);
			LOG( LOG_DEBUG, "[DEBUG] %s thread[%s] TLS sending ans back to client.\n", myname, tname);
			//pthread_mutex_lock(&check_tls_mutex);//lock
//#ifdef DEBUG	
//			cerr << "[DEBUG] thread[" << thread_name << "] successfully acquired check_tls_mutex" << endl;			
//#endif			

#ifdef USE_GNUTLS
			send_len = gnutls_record_send (tls_session, tcp_pkt, ans_size);
#endif //#ifdef USE_GNUTLS

#ifdef USE_OPENSSL
			send_len = SSL_write(ssl, tcp_pkt, ans_size);
			if( send_len <= 0) {
				log_crypto_err("cannot SSL_write");
				exit(-1);
			}
#endif //#ifdef USE_GNUTLS
			if (send_len == -1) 
				die ("thread[recv_response] TLS send returns -1");			
////#ifdef DEBUG	
//			cerr << "[DEBUG] thread[" << thread_name << "] releasing check_tls_mutex" << endl;			
//#endif			
//			pthread_mutex_unlock(&check_tls_mutex);//unlock
			cleanup_pkt(tcp_pkt);
		} else if (my_arg->t == CONN_TCP){//TCP
			tcp_pkt = make_tcp_pkt(ans_pkt, ans_size);
			LOG( LOG_DEBUG, "[DEBUG] %s thread[%s] TCP sending ans back to client.\n", myname, tname);
			send_len = sendpkt(my_arg->s, tcp_pkt, ans_size);
			if (send_len == -1) 
				die ("thread[recv_response] TCP send returns -1");
			cleanup_pkt(tcp_pkt);
		} else if (my_arg->t != CONN_UDP) {  //unknown
			die("thread[recv_response] client connection is unknown.");
		}
		LOG( LOG_DEBUG, "[DEBUG] %s thread[%s] ans has been sent\n", myname, tname);
		//remove the corresponding query from query table
		pthread_mutex_lock(&query_table_mutex);
		if (my_arg->t == CONN_UDP){//UDP, put udp send here is because of the lock above
			if (udp_clnt_map.count(dns_id) != 1) {
				die("query is not found in udp_clnt_map");
			}		
			dns_ident = &(udp_clnt_map[dns_id]);
			put_dns_id(ans_pkt, dns_ident->dns_id);//put the original id
			LOG( LOG_DEBUG, "[DEBUG] %s thread[%s] done changing BACK dns_id, id_before:%d, id_after:%d\n", 
				myname, tname, dns_id, dns_ident->dns_id);				
			send_len = sendpkt_to(my_arg->s, ans_pkt, ans_size, &(dns_ident->clnt_addr), dns_ident->clnt_addr_len);
			if (send_len == -1) 
				die ("thread[recv_response] UDP send returns -1");
			udp_clnt_map.erase(dns_id);//remove it from client table
			dns_ident = NULL;
			proxy_id_list.push_back(dns_id);//make this id available again
		}		
		if (query_table.count(dns_id) != 1) {
			die("query is not found in query table");
		}
		query_table.erase(dns_id);
		if (query_table.size() < MAX_NUM_QUERY) {
			LOG( LOG_DEBUG, "[DEBUG] %s thread[%s] query_table size:%d signal: query_table_cv\n", myname, tname, query_table.size());
			pthread_cond_signal(&query_table_cv);
		}
		pthread_mutex_unlock(&query_table_mutex);

		cleanup_pkt(ans_pkt);
		ans_size = -1;	
	}
	
	//potentailly, the following code will never be excuted because this thread will receive a
	//signal from another thread and terminate
	
	LOG( LOG_DEBUG, "[DEBUG] %s thread[%s] ends\n", myname, tname);
	pthread_exit(NULL);
}

/*
	this function uses two threads to send query and receive response
	when TCP connection from client is close in one (sending) thread, sending thread 
	will send a signal to the (receive) thread to let it exit
	Although handle_tcp and handle_udp are seperated but they might be merged in the future
	
*/

void handle_tcp_client2(int s, string pname)
{
	const char *myname = pname.c_str();
	LOG( LOG_DEBUG, "[DEBUG] %s Main: handle tcp client...\n", myname);

	int rc;
	//definition
	//make all the parameters in array to make it easy for processing
	int NUM_THREADS = 3;
	pthread_t thread[NUM_THREADS];			//array
	void *(*thread_func[NUM_THREADS]) (void *);	//array for function pointers		
	thread_argv_t thread_argv[NUM_THREADS];		//array for thread arguments
	
	thread_func[0] = recv_resp_from_server;
	thread_func[1] = recv_query_from_client;
	thread_func[2] = send_query_to_server;
	
	thread_argv[0].name = "recv_resp_from_server";
	thread_argv[1].name = "recv_query_from_client";
	thread_argv[2].name = "send_query_to_server";
	
	pthread_attr_t attr;
	void *status;
	
	//initialization
	pthread_attr_init(&attr);
	pthread_attr_setdetachstate(&attr, PTHREAD_CREATE_JOINABLE);
	pthread_mutex_init(&query_table_mutex, NULL);
	pthread_mutex_init(&pending_query_mutex, NULL);
	pthread_mutex_init(&check_tls_mutex, NULL);
	
	pthread_cond_init (&query_table_cv, NULL);
	pthread_cond_init (&pending_query_cv, NULL);
	//pthread_cond_init (&tls_session_cv, NULL);
	
	if (dns_socket_udp == -1) {	
		dns_socket_udp = udp_connect (DNS_SERV_PORT, DNS_SERV_IP);//this is the real server
		LOG( LOG_DEBUG, "[DEBUG] %s handle_tcp_client2() udp connect to DNS server... dns_socket_udp:%d\n", myname, dns_socket_udp);		
	}
	
	for (int i=0; i<NUM_THREADS; i++) {
		LOG( LOG_DEBUG, "[DEBUG] %s Main: creating thread[%d]\n", myname, i);

		//if (i == 1)
		thread_argv[i].s = s;
		thread_argv[i].t = CONN_TCP;
		thread_argv[i].pname = pname;
		rc = pthread_create(&thread[i], &attr, thread_func[i], (void *)&thread_argv[i]);
		if (rc) {
			cerr << "[ERROR] " << pname << "Main: return code from pthread_create() is " << rc << endl;
			exit(-1);			
		}
		
		thread_map[thread[i]] = thread_argv[i].name;
		if (thread_argv[i].name == "recv_resp_from_server") {
			LOG( LOG_DEBUG, "[DEBUG] Main: Created thread recv_response:%u\n", thread[i]);
			recv_resp_thread = &thread[i];		
		} else if (thread_argv[i].name == "send_query_to_server"){
			LOG( LOG_DEBUG, "[DEBUG] Main: Created thread send_query:%u\n", thread[i]);
			send_query_thread = &thread[i];				
		}	
	}
	pthread_attr_destroy(&attr);
	for (int i=0; i<NUM_THREADS; i++) {
		rc = pthread_join(thread[i], &status);
		if (rc) {
			cerr << "[ERROR] " << pname << "Main: return code from pthread_join() is " << rc << endl;
			exit(-1);
		}
		LOG( LOG_DEBUG, "[DEBUG] %s Main: completed join with thread %d, status:%d\n", myname, i, (long)status);
	}

	//clean up
	LOG( LOG_DEBUG, "[DEBUG] %s all threads are gone and doing clean up\n", myname);
	pthread_mutex_destroy(&query_table_mutex);
	pthread_mutex_destroy(&pending_query_mutex);
	pthread_mutex_destroy(&check_tls_mutex);
	pthread_cond_destroy(&query_table_cv);
	pthread_cond_destroy(&pending_query_cv);
	//pthread_cond_destroy(&tls_session_cv);
	shutdown (s, SHUT_RDWR); 
	close(s);
}

/*
	this would be similar to handle_tcp_client2 which will use threads
*/
void handle_udp_client2(int s, string pname)
{
	const char *myname = pname.c_str();
	LOG( LOG_DEBUG, "[DEBUG] %s Main: handle UDP client...\n", myname);
	
	int rc;
	//definition
	//make all the parameters in array to make it easy for processing
	int NUM_THREADS = 3;
	pthread_t thread[NUM_THREADS];			//array
	void *(*thread_func[NUM_THREADS]) (void *);	//array for function pointers		
	thread_argv_t thread_argv[NUM_THREADS];		//array for thread arguments
	
	thread_func[0] = recv_resp_from_server;
	thread_func[1] = recv_query_from_client;
	thread_func[2] = send_query_to_server;
	
	thread_argv[0].name = "recv_resp_from_server";
	thread_argv[1].name = "recv_query_from_client";
	thread_argv[2].name = "send_query_to_server";
	
	pthread_attr_t attr;
	void *status;
	
	//initialization
	pthread_attr_init(&attr);
	pthread_attr_setdetachstate(&attr, PTHREAD_CREATE_JOINABLE);
	pthread_mutex_init(&query_table_mutex, NULL);
	pthread_mutex_init(&pending_query_mutex, NULL);
	pthread_mutex_init(&check_tls_mutex, NULL);
	pthread_mutex_init(&udp_clnt_map_mutex, NULL);
	
	pthread_cond_init (&query_table_cv, NULL);
	pthread_cond_init (&pending_query_cv, NULL);
	//pthread_cond_init (&tls_session_cv, NULL);
	
	for (int i=1; i<= 65535; i++) {
		proxy_id_list.push_back(i);
	}
	
	if (dns_socket_udp == -1) {	
		dns_socket_udp = udp_connect (DNS_SERV_PORT, DNS_SERV_IP);//this is the real server
		LOG( LOG_DEBUG, "[DEBUG] %s handle_tcp_client2() udp connect to DNS server... dns_socket_udp: %d\n", myname, dns_socket_udp);
	}
	
	for (int i=0; i<NUM_THREADS; i++) {
		LOG( LOG_DEBUG, "[DEBUG] %s Main: creating thread[%d]\n", myname, i);
		
		//if (i == 1)
		thread_argv[i].s = s;
		thread_argv[i].t = CONN_UDP;
		thread_argv[i].pname = pname;
		rc = pthread_create(&thread[i], &attr, thread_func[i], (void *)&thread_argv[i]);
		if (rc) {
			cerr << "[ERROR] " << pname << "Main: return code from pthread_create() is " << rc << endl;
			exit(-1);			
		}
		
		thread_map[thread[i]] = thread_argv[i].name;
		if (thread_argv[i].name == "recv_resp_from_server") {
			LOG( LOG_DEBUG, "[DEBUG] Main: Created thread recv_response:%u\n", thread[i]);
			recv_resp_thread = &thread[i];		
		} else if (thread_argv[i].name == "send_query_to_server"){
			LOG( LOG_DEBUG, "[DEBUG] Main: Created thread send_query:%u\n", thread[i]);
			send_query_thread = &thread[i];				
		}	
	}
	pthread_attr_destroy(&attr);
	for (int i=0; i<NUM_THREADS; i++) {
		rc = pthread_join(thread[i], &status);
		if (rc) {
			cerr << "[ERROR] " << pname << "Main: return code from pthread_join() is " << rc << endl;
			exit(-1);
		}
		LOG( LOG_DEBUG, "[DEBUG] %s Main: completed join with thread %d, status:%d\n", myname, i, (long)status);
	}

	//clean up
	LOG( LOG_DEBUG, "[DEBUG] %s all threads are gone and doing clean up\n", myname);
	pthread_mutex_destroy(&query_table_mutex);
	pthread_mutex_destroy(&pending_query_mutex);
	pthread_mutex_destroy(&check_tls_mutex);
	pthread_mutex_destroy(&udp_clnt_map_mutex);
	pthread_cond_destroy(&query_table_cv);
	pthread_cond_destroy(&pending_query_cv);
	//pthread_cond_destroy(&tls_session_cv);
	//shutdown (s, SHUT_RDWR); 
	close(s);
}


/*
	this simply recv UDP query, send to server and send the response back to server
*/
void handle_udp_client(int s)
{
	int send_len = -1;
	int recv_len = -1;
	int query_size = -1;
	int ans_size = -1;
	unsigned char* query_pkt = NULL;
	unsigned char* ans_pkt = NULL;	
	struct sockaddr client_addr;
	socklen_t client_addr_len = sizeof(client_addr);
	struct sockaddr serv_addr;
	socklen_t serv_addr_len = sizeof(serv_addr);
	query_pkt = udp_pull_dns_pkt(s, query_size, &client_addr, &client_addr_len);
	if (dns_socket_udp == -1) {	
		dns_socket_udp = udp_connect (DNS_SERV_PORT, DNS_SERV_IP);//this is the real server
#ifdef DEBUG	
		cerr << "[DEBUG] udp connect to DNS server... dns_socket_udp:" << dns_socket_udp << endl;
#endif			
	}
	send_len = sendpkt(dns_socket_udp, query_pkt, query_size);					
	//cout<<"send_len="<<send_len<<endl;
	ans_pkt = udp_pull_dns_pkt(dns_socket_udp, ans_size, &serv_addr, &serv_addr_len);
#ifdef DEBUG	
	cerr << "[DEBUG] receive the ans from dns_socket_udp:" << dns_socket_udp << endl;
#endif						
	print_dns_msg(ans_pkt, ans_size);
	send_len = sendpkt_to(s, ans_pkt, ans_size, &client_addr, client_addr_len);
}

int start_proxy(int port_num)
{
	cerr << "[info] start_front_server: port_num is " << port_num << endl;
        signal (SIGQUIT, sig_handler);
        signal (SIGINT, sig_handler);
        signal (SIGALRM, sigthread_handler);
        //--------------------------------------------------
        //create UDP socket
        //--------------------------------------------------
        struct sockaddr_in udp_addr;
        int udp_socket = 0;
        if ((udp_socket = socket(AF_INET,SOCK_DGRAM,0)) == -1) {
                perror("udp_socket");
                exit(1);
        }

        memset(&udp_addr,0,sizeof(struct sockaddr_in));
        udp_addr.sin_family = AF_INET;
        udp_addr.sin_addr.s_addr = htonl(INADDR_ANY);
        udp_addr.sin_port = htons(port_num);

        if (bind(udp_socket,(struct sockaddr *)&udp_addr,sizeof(udp_addr)) == -1) {
                perror("udp_bind");
                exit(1);
        }
        sockets.push_back(udp_socket);
        //--------------------------------------------------
        //create UDP socket ends
        //-------------------------------------------------- 	
	
        //--------------------------------------------------
        //create TCP socket
        //-------------------------------------------------- 
        int tcp_socket = 0;
        if ((tcp_socket = socket(AF_INET,SOCK_STREAM,0)) == -1) {
                perror("socket");//AF_INET => IPv4, SOCK_STREAM=>stream socket(TCP)     
                exit(1);
        }

        int yes = 1;
        if ( setsockopt(tcp_socket, SOL_SOCKET, SO_REUSEADDR, &yes, sizeof(int)) == -1 )
        {
              perror("setsockopt");
        }

        struct sockaddr_in tcp_addr;
        memset(&tcp_addr,0,sizeof(struct sockaddr_in));
        tcp_addr.sin_family = AF_INET;
        tcp_addr.sin_addr.s_addr = htonl(INADDR_ANY);
        tcp_addr.sin_port = htons(port_num); //let the server find a port number
   
        //bind
        if (bind(tcp_socket,(struct sockaddr *)&tcp_addr,sizeof(tcp_addr)) == -1) {
                perror("bind");
                exit(1);
        }

        //listen
        int backlog = 1024;//how many connection ready to be queued
        if (listen(tcp_socket, backlog) == -1) {
                perror("listen");
                exit(1);
        }
        sockets.push_back(tcp_socket);
        //--------------------------------------------------
        //create TCP socket ends
        //-------------------------------------------------- 
        
        //--------------------------------------------------
        //select socket
        //-------------------------------------------------- 
        fd_set rset;
        FD_ZERO(&rset);
        int max_fd = UTIL_MAX(tcp_socket, udp_socket) + 1;
        int status_t;
        pid_t pid;
        
        struct timeval	timeout;
        timeout.tv_sec  = 10;
        timeout.tv_usec = 0;
        
        struct sockaddr_in clnt_addr;  //for accepted client
        int clnt_len = sizeof(clnt_addr);
        int clnt_socket = -1;
        struct sigaction sa;
        
        sa.sa_handler = sigchld_handler; // reap all dead processes (from beej)
        sigemptyset(&sa.sa_mask);
        sa.sa_flags = SA_RESTART;
        if (sigaction(SIGCHLD, &sa, NULL) == -1) {
        	perror("sigaction");
		exit(1);
	}
	
        while (1) {
        	cerr << "[info] waiting for connection..." << endl;
                //select_restart:
                FD_SET(tcp_socket, &rset);
                FD_SET(udp_socket, &rset);
                status_t = select(max_fd, &rset, NULL, NULL, NULL);
                if (status_t < 0) {
                	if (errno == EINTR) {// some signal just interrupted us, so restart
                		continue;
                	//goto select_restart;
                	}                
                        perror("select");
                        exit(1);
                }
                if (status_t == 0) {
                	cerr << "time out"<<endl;
                        continue;
                }
                if (FD_ISSET(tcp_socket, &rset)) { //tcp
                	cerr << "[info] TCP connection from client..."<<endl;
                	clnt_socket = accept(tcp_socket, (struct sockaddr *)(&clnt_addr),(socklen_t *)&clnt_len);
                	pid = fork();
                	if (pid < 0) {
                		die("fork: pid < 0");
                	} else if (pid == 0) { //child
                		close(tcp_socket);
                		close(udp_socket);
                		//handle_tcp_client(clnt_socket);
                		handle_tcp_client2(clnt_socket, "");
                		//handle_tcp_client3(clnt_socket);
                		cerr << "[info] child exit..."<<endl;
                		exit(0);
                	} else { //parent
                		cerr << "[info] parent continue..."<<endl;
                		close(clnt_socket); 
                		//childProcCount++;
                		continue;
                	}
                }
                if (FD_ISSET(udp_socket, &rset)) { //udp
                        cerr << "[info] UDP connection from client..."<<endl;                     
                        handle_udp_client(udp_socket);
                }

        }
        shutdown (tcp_socket, SHUT_RDWR); 
        close(tcp_socket);
        close(udp_socket);
	return 0;
}

/*
	tcp process, fork child for each incoming TCP connection
*/
void tcp_process(int port_num, string pname)
{
	const char *myname = pname.c_str();
#ifdef USE_OPENSSL
	LOG( LOG_DEBUG, "[DEBUG] %s initial openssl\n",myname);
	ERR_load_SSL_strings();
	OpenSSL_add_all_algorithms();
	SSL_library_init();
	LOG( LOG_DEBUG, "[DEBUG] %s DONE initial openssl\n", myname);
#endif

#ifdef USE_GNUTLS
	LOG( LOG_DEBUG, "[DEBUG] %s initial gnutls\n",myname);
	gnutls_global_init();
	generate_dh_params();//this will be very slow
	LOG( LOG_DEBUG, "[DEBUG] %s DONE initial gnutls\n", myname);
#endif
	LOG( LOG_DEBUG, "[DEBUG] %s is running\n", myname);
       //--------------------------------------------------
        //create TCP socket
        //-------------------------------------------------- 
	int tcp_socket = -1;
        if ((tcp_socket = socket(AF_INET,SOCK_STREAM,0)) == -1) {
                perror("socket");//AF_INET => IPv4, SOCK_STREAM=>stream socket(TCP)     
                exit(1);
        }

        int yes = 1;
        if (setsockopt(tcp_socket, SOL_SOCKET, SO_REUSEADDR, &yes, sizeof(int)) == -1) {
              perror("setsockopt");
        }

        struct sockaddr_in tcp_addr;
        memset(&tcp_addr,0,sizeof(struct sockaddr_in));
        tcp_addr.sin_family = AF_INET;
        tcp_addr.sin_addr.s_addr = htonl(INADDR_ANY);
        tcp_addr.sin_port = htons(port_num);
   
        //bind
        if (bind(tcp_socket,(struct sockaddr *)&tcp_addr,sizeof(tcp_addr)) == -1) {
                perror("bind");
                exit(1);
        }

        //enable TCP fast open
        int flag = 1;
        int result = setsockopt(tcp_socket,            /* socket affected */
                              IPPROTO_TCP,     /* set option at TCP level */
                              TCP_FASTOPEN,     /* name of option */
                              (char *) &flag,  /* the cast is historical cruft */
                              sizeof(int));    /* length of option value */
    
        if(result < 0)
        {   
          perror("enable TCP fastopen failed\n");
          exit(1);
        }   
        else
        {   
          printf("TCP fastopen enabled\n");
        }   


        //listen
        int backlog = 1024;//how many connection ready to be queued
        if (listen(tcp_socket, backlog) == -1) {
                perror("listen");
                exit(1);
        }
        sockets.push_back(tcp_socket);
        //--------------------------------------------------
        //create TCP socket ends
        //-------------------------------------------------- 
        
        //--------------------------------------------------
        //loop forever until break
        //--------------------------------------------------
        struct sockaddr_in clnt_addr;  //for accepted client
        int clnt_len = sizeof(clnt_addr);
        int clnt_socket = -1;
        pid_t pid;    
        struct sigaction sa;
        sa.sa_handler = sigchld_handler; // reap all dead processes (from beej)
        sigemptyset(&sa.sa_mask);
        sa.sa_flags = SA_RESTART;
        if (sigaction(SIGCHLD, &sa, NULL) == -1) {
        	perror("sigaction");
		exit(1);
	}        
        while (1) {
        	cerr << "[info] " << pname << " waiting for connection..."<<endl;
		clnt_socket = accept(tcp_socket, (struct sockaddr *)(&clnt_addr),(socklen_t *)&clnt_len);
		//sockets.push_back(clnt_socket);
		pid = fork();
		if (pid < 0) {
			die("fork: pid < 0");
		} else if (pid == 0) { //child
			close(tcp_socket);
			handle_tcp_client2(clnt_socket, pname);
                	cerr << "[info] " << pname << " child exit..."<<endl;
                	exit(0);
                } else { //parent
                	cerr << "[info] " << pname << " parent continue..."<<endl;
                	close(clnt_socket);
                	continue;
                }        
        }
	close(tcp_socket);
	LOG( LOG_DEBUG, "[DEBUG] %s ends\n", myname);
#ifdef USE_GNUTLS
	gnutls_global_deinit();
#endif //USE_GNUTLS
}

/*
	udp process
*/
void udp_process(int port_num, string pname)
{
	const char *myname = pname.c_str();
	LOG( LOG_DEBUG, "[DEBUG] %s is running\n", myname);

	//--------------------------------------------------
        //create UDP socket
        //--------------------------------------------------
        struct sockaddr_in udp_addr;
        int udp_socket = -1;
        if ((udp_socket = socket(AF_INET,SOCK_DGRAM,0)) == -1) {
                perror("udp_socket");
                exit(1);
        }

        memset(&udp_addr,0,sizeof(struct sockaddr_in));
        udp_addr.sin_family = AF_INET;
        udp_addr.sin_addr.s_addr = htonl(INADDR_ANY);
        udp_addr.sin_port = htons(port_num);

        if (bind(udp_socket,(struct sockaddr *)&udp_addr,sizeof(udp_addr)) == -1) {
                perror("udp_bind");
                exit(1);
        }
        //--------------------------------------------------
        //create UDP socket ends
        //--------------------------------------------------
        
        //do_udp
        handle_udp_client2(udp_socket, pname);
	LOG( LOG_DEBUG, "[DEBUG] %s ends\n", myname);
}

/*
	this will fork udp and tcp as seperated process and 
*/
int start_proxy2(int port_num)
{
	if (MAX_NUM_QUERY > 65535) {
		die("MAX_NUM_QUERY > 65535");
	}
	cerr << "[info] Process[Main] start proxy: port_num is " << port_num << endl;
	cerr << "[info] Process[Main] UDP and TCP will be forked as seperated processes" << endl;
	signal (SIGQUIT, sig_handler);
	signal (SIGINT, sig_handler);
	signal (SIGALRM, sigthread_handler);
	//signal (SIGCHLD, sigchld_handler);
	int NUM_PROCESSES = 2;
	void (*thread_func[NUM_PROCESSES]) (int, string);	//array for function pointers			
	thread_func[0] = &tcp_process;
	thread_func[1] = &udp_process;
	string process_name[NUM_PROCESSES];
	process_name[0] = "Process[TCP]";
	process_name[1] = "Process[UDP]";
	
	pid_t child_pid, wpid;
	for (int i=0; i<2; i++) {
		child_pid = fork();
		if (child_pid < 0) {
			die("fork: pid < 0");
		} else if (child_pid == 0) { //child
			cerr << "[info] Process[Main] do " << process_name[i] << " ..." << endl;
			(*thread_func[i])(port_num, process_name[i]);
			cerr << "[info] Process[Main] done " << process_name[i] << " ..." << endl;
			exit(0);
        	} 
#ifdef DEBUG	
		else { //parent
	        	cerr << "[DEBUG] Process[Main] parent continues"<< endl;	
			continue; //nothing
        	}
#endif        			
	}

	LOG( LOG_DEBUG, "[DEBUG] Process[Main] waiting for child processes\n");
	int status = 0;
	while ((wpid = wait(&status)) > 0) {
		cout << "[info] Process[Main] Exit status of [" << (int)wpid << "] is " << status << endl;
	}
	cout << "[info] Process[Main] ends" << endl;
        return 0;
}

